<?php
	
	declare(strict_types=1);
	
	/**
	 * @author: Daniel Schmitzer (daniel@maennchen1.de)
	 * @date: 20.10.23
	 * @time: 14:48
	 */
	
	namespace wpsgTicket;
	
	class Ticket {
	
		private static array $arInstances = [];
		
		private ?int $id = null;
		private ?array $data = null;
	
		public const STATE_ACTIVE = 1;
		public const STATE_INACTIVE = 2;
		public const STATE_DELETED = 3;
		
		public const BOOKING_STATE_OPEN = 1;
		public const BOOKING_STATE_CLOSED = 2;
		
		private function __construct($id) {
		
			global $wpdb;
			
			$this->id = $id;
			
			$query = $wpdb->prepare("SELECT * FROM `".$wpdb->prefix."wpsg_tickets` WHERE id = %d", $id);
			$this->data = $wpdb->get_row($query, ARRAY_A);
			
			if (intval($id) !== intval($this->data['id']??-1)) throw new \Exception(sprintf(__('Ticket mit ID %s kann nicht geladen werden', 'wpsg'), $id));
			
		}
		
		public function getStartDate(): \DateTime {
			
			$start_date_time = $this->__get('from_date');
			
			if ($this->__get('from_time') !== '') $start_date_time .= ' '.$this->__get('from_time');
			
			return new \DateTime($start_date_time);
			
		}
		
		public function getEndDate(): \DateTime {
		
			if ($this->__get('to_date') !== '') return new \DateTime($this->__get('to_date'));
			else return $this->getStartDate();
		
		}
		
		public function getDates(): array {
			
			return Plugin::getDatesBetween(
				$this->getStartDate()->format('Y-m-d'), $this->getEndDate()->format('Y-m-d')
			);
			
		}
		
		public function getProductID(): int {
			
			return intval($this->__get('p_id'));
			
		}
		
		public function getProduct(): \wpsg_product {
			
			return \wpsg_product::getInstance($this->getProductID());
			
		}
		
		/**
		 * Ist das Ticket an Datum $td verfügbar
		 *
		 * @param \DateTime $td
		 * @return void
		 */
		public function isActive(\DateTime $td) {
		
			return in_array($td->format('Y-m-d'), $this->getDates());
		
		}
		
		public function isAvailable(): bool {
		
			if ($this->getFreeSlotCount() <= 0) return false;
			
			if ($this->getState() === self::STATE_INACTIVE || $this->getState() === self::STATE_DELETED) return false;
			
			// Vergangene Termine ausblenden
			if ($this->getStartDate() < new \DateTime()) {
				
				if ($this->__get('from_time') === '' && $this->__get('from_date') === (new \DateTime())->format('Y-m-d')) return true;
				else return false;
			
			}
			
			return true;
		
		}
		
		public function getId(): int {
			
			return $this->id;
			
		}
		
		public function changeState(): void {
			
			global $wpdb;
			
			$new_state = null;
			
			if ($this->getState() === self::STATE_ACTIVE) $new_state = self::STATE_INACTIVE;
			else if ($this->getState() === self::STATE_INACTIVE) $new_state = self::STATE_ACTIVE;
			
			if ($new_state !== null) {
				
				$wpdb->update(
					$wpdb->prefix.'wpsg_tickets',
					[ 'state' => $new_state ],
					[ 'id' => $this->getId() ],
					[ '%d' ],
					[ '%d' ]
				);
				
				$this->data['state'] = strval($new_state);
			
			}
			
		}
		
		public function delete(): void {
			
			global $wpdb;
			
			$wpdb->update(
				$wpdb->prefix.'wpsg_tickets',
				[ 'state' => self::STATE_DELETED ],
				[ 'id' => $this->getId() ],
				[ '%d' ],
				[ '%d' ]
			);
			
			if ($this->isSerial()) {
				
				foreach (self::find(['autoticket_id' => $this->getId()], null, true) as $oTicket) {
					
					$oTicket->delete();
					
				}
				
			}
			
			$this->data['state'] = strval(self::STATE_DELETED);
			
		}
		
		public function getSerialTicket(): Ticket {
			
			return Ticket::getInstance(intval($this->__get('autoticket_id')));
			
		}
		
		/**
		 * Gibt true zurück, wenn der Termin aus einem Serientermin erstellt wurde
		 * @return bool
		 */
		public function isSerialTicket(): bool {
			
			return intval($this->__get('autoticket_id')) > 0;
			
		}
		
		/**
		 * Gibt true bei einem Serientermin zurück
		 * @return bool
		 */
		public function isSerial(): bool {
			
			return $this->__get('autoticket') !== '';
			
		}
		
		public function getAutoConfig(): array {
			
			$autoconfig = $this->__get('autoticket');
			$autoconfig = json_decode($autoconfig, true);

			if (!is_array($autoconfig)) throw new \Exception(__('Konfiguration des Serientermins kann nicht gelesen werden!'));
			
			return $autoconfig;
			
		}
		
		public function createTickets(): void {
			
			global $wpdb;
		
			if ($this->isSerial()) {
			
				$conf = $this->getAutoConfig();
				
				$from_date_obj = new \DateTime($conf['from_date']);
				$to_date_obj = new \DateTime($conf['to_date']);
				
				$interval = new \DateInterval('P1D');
				$dates = new \DatePeriod($from_date_obj, $interval, $to_date_obj->modify('+1 day'));
 
				$today = new \DateTime();
				
				foreach ($dates as $date) {

					if ($date < $today) continue;
					if ($date->diff($today)->days > intval($conf['days_future']??0)) continue;
					
					$strDay = $date->format('Y-m-d');
					$weekday = (intval($date->format('w')) + 6) % 7;
					
					if (in_array($weekday, $conf['days'])) {
					
						$where_condition = []; foreach([
							'autoticket_id' => $this->getId(),
							'from_date' => $strDay,
							'to_date' => $strDay
						] as $column => $value) $where_condition[] = $wpdb->prepare("$column = %s", $value);
						$query = "SELECT * FROM ".WPSG_TBL_TICKETS." WHERE ".implode(' AND ', $where_condition);
						$result = $wpdb->get_row($query);
						
						if (!$result) {
						
							$wpdb->insert(WPSG_TBL_TICKETS, [
								'p_id' => $this->getProductID(),
								'from_date' => $strDay,
								'from_time' => $conf['event_data']['from_time'],
								'to_date' => $strDay,
								'to_time' => $conf['event_data']['to_time'],
								'slots' => $conf['event_data']['slots'],
								'state' => self::STATE_ACTIVE,
								'autoticket' => '',
								'autoticket_id' => $this->getId()
							]);
							
						}
						
					}
					
				}
			
			} else throw new \Exception(__('Aufrufsfehler!'.$this->getId(), 'wpsg'));
		
		}
		
		public function getTicketPublicURL(int $order_product_id, int $amount_index): string {
			
			//$oTicketBooking = TicketBooking::getInstanceByBookingKey(TicketBooking::encodeBookingKey($order_product_id, ))
			
			$oOrderProduct = \wpsg_order_product::getInstance($order_product_id);
			
			$arTicketSecrets = json_decode($oOrderProduct->getMeta('ticket_secrets'));
	 
			return \add_query_arg([
				'action' => 'wpsg_ticket_frontend',
				'ticket_action' => 'get',
				'order_product_id' => $oOrderProduct->getId(),
				'amount_index' => $amount_index,
				'hash' => hash_hmac('sha256', strval($oOrderProduct->getId().$amount_index.$this->getId()), $arTicketSecrets[$amount_index])
			], \admin_url('admin-ajax.php'));
			
		}
		
		public function toJSON(): array {
			
			$r = [
				'id' => $this->id,
				'autoticket_id' => intval($this->__get('autoticket_id')),
				'label' => $this->getLabel(),
				'label_short' => $this->getShortLabel(),
				'slots' => $this->getSlotCount(),
				'slots_used' => $this->getUsedSlotCount(),
				'slots_available' => $this->getFreeSlotCount(),
				'state' => $this->getState(),
				'booking_state' => $this->getBookingState(),
				'dates' => $this->getDates(),
				'amount' => 1,
				'order' => $this->getOrderInfo()
			];
			
			if ($this->isSerial()) {
				
				$r['is_autoticket'] = true;
				$r['autoticket'] = json_decode($this->__get('autoticket'));
				$r['count_tickets'] = self::count(['autoticket_id' => $this->getId()]);
				
			} else {
				
				$r['is_autoticket'] = false;
				
			}
			
			return $r;
			
		}
		
		public function getState(): int { return intval($this->__get('state')); }
		
		public function getOrderInfo(): array {
			
			global $wpdb;
			
			$arOrderID = $wpdb->get_col(
				$wpdb->prepare("
					SELECT
    					DISTINCT O.`id`
					FROM
						`".WPSG_TBL_META."` AS M
							LEFT JOIN `".WPSG_TBL_ORDERPRODUCT."` AS OP ON (M.`target_id` = OP.`id`)
							LEFT JOIN `".WPSG_TBL_ORDER."` AS O ON (OP.`o_id` = O.`id`)
					WHERE
						M.`meta_table` = 'WPSG_TBL_ORDERPRODUCT' AND
						M.`meta_key` = 'ticket_id' AND
						M.`meta_value` = %d AND
						O.`status` != %d
					", $this->getId(), \wpsg_ShopController::STATUS_STORNIERT
				)
			);
			
			$arReturn = [];
			
			foreach ($arOrderID as $order_id) {
				
				$arReturn[] = [
					'amount' => intval($wpdb->get_var(
						$wpdb->prepare("
							SELECT
							    SUM(OP.`menge`)
							FROM
								 `".WPSG_TBL_ORDERPRODUCT."` AS OP
									LEFT JOIN `".WPSG_TBL_META."` AS M ON (M.`target_id` = OP.`id` AND M.`meta_table` = 'WPSG_TBL_ORDERPRODUCT' AND M.`meta_key` = 'ticket_id')
							WHERE
								OP.`o_id` = %d AND
								M.`meta_value` = %d
						", $order_id, $this->getId())
					)),
					'order_id' => intval($order_id),
					'order_nr' => \wpsg_order::getInstance($order_id)->getNr(),
					'order_invoice_name' => \wpsg_order::getInstance($order_id)->getInvoiceFirstName().' '.\wpsg_order::getInstance($order_id)->getInvoiceName(),
					'order_link' => \wpsg_admin_url('Order', 'view', ['edit_id' => $order_id], [], true)
				];
				
			}
			
			return $arReturn;
			
		}
		
		public function getBookingState(): int {
			
			if (!$this->isAvailable()) return self::BOOKING_STATE_CLOSED;
			
			return self::BOOKING_STATE_OPEN;
			
		}
		
		public function getFreeSlotCount(): int { return $this->getSlotCount() - $this->getUsedSlotCount(); }
		
		public function getSlotCount(): int { return intval($this->__get('slots')); }
		
		public function getUsedSlotCount(): int {
			
			global $wpdb;
			
			return intval($wpdb->get_var(
				$wpdb->prepare("
					SELECT
						SUM(OP.`menge`)
					FROM
						`".WPSG_TBL_META."` AS M
							LEFT JOIN `".WPSG_TBL_ORDERPRODUCT."` AS OP ON (OP.`id` = M.`target_id`)
					WHERE
						M.`meta_table` = 'WPSG_TBL_ORDERPRODUCT' AND
						M.`meta_key` = 'ticket_id' AND
						M.`meta_value` = %d AND
						OP.`id` IS NOT NULL
				", $this->getId())
			));
			
		}
		
		public function getShortLabel(): string {
			
			if ($this->isSerial()) {
			
				return $this->__get('label_short');
				
			} else return '';
			
		}
		
		public function getLabel(): string {
			
			if ($this->isSerial()) {
				
				if (trim($this->__get('label')) !== '') return $this->__get('label');
				else return __('Serientermin #', 'wpsg').$this->getId();
				
			}
			
			$strLabel = date('d.m.Y', strtotime($this->__get('from_date')));
			
			if ($this->__get('from_time') !== '') $strLabel .= ' '.self::clearTime($this->__get('from_time'));
			
			if ($this->__get('to_date') !== '') {
				
				$strLabel .= ' - ';
				
				if ($this->__get('from_date') !== $this->__get('to_date')) $strLabel .= date('d.m.Y', strtotime($this->__get('to_date'))).' ';
				
				if ($this->__get('to_time') !== '') $strLabel .= self::clearTime($this->__get('to_time'));
				
			}
			
			return $strLabel;
			
		}
		
		public function __get($key): string {
			
			return $this->data[$key]??'';
			
		}
		
		public static function getQueryWHERE(array $arFilter): string {
			
			global $wpdb;
			
			$strQueryWHERE = "";
	 
			if (intval($arFilter['product_id']??-1) > -1) $strQueryWHERE .= $wpdb->prepare(" AND `p_id` = %d ", intval($arFilter['product_id']));
			if (intval($arFilter['autoticket_id']??-1) > 0) $strQueryWHERE .= $wpdb->prepare(" AND `autoticket_id` = %d ", intval($arFilter['autoticket_id']));
			
			if (intval($arFilter['state']) > 0) $strQueryWHERE .= $wpdb->prepare(" AND `state` = %d ", intval($arFilter['state']));
			else if (intval($arFilter['state']) !== -1) $strQueryWHERE .= $wpdb->prepare(" AND `state` != %d ", intval(self::STATE_DELETED));
	 	
			if (preg_match('/^\d{4}\-\d{2}\-\d{2}$/', $arFilter['day']??'')) {
			
				$strQueryWHERE .= "AND
					(
						(`to_date` IS NULL AND `from_date` = CAST('".wpsg_q($arFilter['day'])."' AS DATE)) OR
						(`to_date` >= CAST('".wpsg_q($arFilter['day'])."' AS DATE) AND `from_date` <= CAST('".wpsg_q($arFilter['day'])."' AS DATE))
					)
				";
			
			}
			
			return $strQueryWHERE;
			
		}
		
		public static function count(array $arFilter): int {
			
			global $wpdb;
			
			$strQueryWHERE = self::getQueryWHERE($arFilter);
			
			$arID = $wpdb->get_col("
				SELECT
					`id`
				FROM
					`".$wpdb->prefix."wpsg_tickets`
				WHERE
					1
					".$strQueryWHERE." AND COALESCE(`autoticket`, '') = ''
			");
			
			$return = function(array $arID) use ($arFilter): int {
				
				if (($arFilter['hideInactive']??false) === true) {
					
					foreach ($arID as $k => $v) {
					
						$o = Ticket::getInstance(intval($v));
						
						if (!$o->isAvailable()) unset($arID[$k]);
						
					}
					
				}

				return sizeof($arID);
				
			};
			
			return $return($arID);
			
		}
		
		/**
		 * @param array $arFilter
		 * @param array|null $arPager
		 * @param bool $getObject
		 * @param $getJSON
		 * @return Ticket[]
		 */
		public static function find(array $arFilter, ?array $arPager = null, bool $getObject = false, $getJSON = false): array {
		
			global $wpdb;
			
			$strQueryWHERE = self::getQueryWHERE($arFilter);
			
			$arID = $wpdb->get_col("
				SELECT
					`id`
				FROM
					`".$wpdb->prefix."wpsg_tickets`
				WHERE
					1
					".$strQueryWHERE." AND COALESCE(`autoticket`, '') = ''
			");
			
			if (($arFilter['hideInactive']??false) === true) {
			
				foreach ($arID as $k => $id) {
					
					$o = Ticket::getInstance(intval($id));
					
					if (!$o->isAvailable()) unset($arID[$k]);
					
				}
				
			}
			
			$arReturn = $arID;
			
			if ($arPager !== null) {
				
				$per_page = intval($arPager['per_page'] ?? 1);
				$page = intval($arPager['page'] ?? 1);
				
				$offset = ($page - 1) * $per_page;
				
				$arReturn = array_slice($arReturn, $offset, $per_page);
				
			}
			
			if ($getObject === true) {
				
				foreach ($arReturn as $k => $v) $arReturn[$k] = Ticket::getInstance(intval($v));
				
			} else if ($getJSON === true) {
				
				foreach ($arReturn as $k => $v) $arReturn[$k] = Ticket::getInstance(intval($v))->toJSON();
				
			}
			
			return array_values($arReturn);
			
		}
		
		public static function countSerial(array $arFilter): int {
			
			global $wpdb;
			
			$strQueryWHERE = self::getQueryWHERE($arFilter);
			
			$arID = $wpdb->get_col("
				SELECT
					`id`
				FROM
					`".$wpdb->prefix."wpsg_tickets`
				WHERE
					1
					".$strQueryWHERE." AND COALESCE(`autoticket`, '') != ''
			");
			
			$return = function(array $arID) use ($arFilter): int {
				
				if (($arFilter['hideInactive']??false) === true) {
					
					foreach ($arID as $k => $v) {
					
						$o = Ticket::getInstance(intval($v));
						
						if (!$o->isAvailable()) unset($arID[$k]);
						
					}
					
				}

				return sizeof($arID);
				
			};
			
			return $return($arID);
			
		}
		
		/**
		 * @param array $arFilter
		 * @param bool $getObject
		 * @param bool $getJSON
		 * @return Ticket[]
		 */
		public static function findSerial(array $arFilter, bool $getObject = false, bool $getJSON = false): array {
			
			global $wpdb;
			
			$strQueryWHERE = self::getQueryWHERE($arFilter);
			
			$arID = $wpdb->get_col("
				SELECT
					`id`
				FROM
					`".$wpdb->prefix."wpsg_tickets`
				WHERE
					1
					".$strQueryWHERE." AND COALESCE(`autoticket`, '') != ''
			");
			
			if (($arFilter['hideInactive']??false) === true) {
			
				foreach ($arID as $k => $id) {
					
					$o = Ticket::getInstance(intval($id));
					
					if (!$o->isAvailable()) unset($arID[$k]);
					
				}
				
			}
			
			$arReturn = $arID;
			
			if ($getObject === true) {
				
				foreach ($arReturn as $k => $v) $arReturn[$k] = Ticket::getInstance(intval($v));
				
			} else if ($getJSON === true) {
				
				foreach ($arReturn as $k => $v) $arReturn[$k] = Ticket::getInstance(intval($v))->toJSON();
				
			}
			
			return array_values($arReturn);
			
		}
		
		public static function getInstance($id): Ticket {
   
			if (!isset(self::$arInstances[$id])) self::$arInstances[$id] = new self($id);
			
        	return self::$arInstances[$id];
    
		}
		
		public static function clearTime(string $time): string {
			
			return preg_replace('/\:00$/', '', $time);
			
		}
		
	}