<?php
	
	declare(strict_types=1);
	
	/**
	 * @author: Daniel Schmitzer (daniel@maennchen1.de)
	 * @date: 18.11.23
	 * @time: 15:58
	 */
	
	namespace wpsgTicket;
	
	class TicketBooking {
	 
		const LOG_SCAN = 1;
		const LOG_REVOKE = 2;
		
		const STATUS_AVAILABLE = 1;
		const STATUS_UNAVAILABLE = 2;
		
		private static array $arInstances = array();
		
		private ?int $booking_id = null;
		public ?array $data = null;
		
		public function __construct(int $booking_id) {
		
			global $wpdb;
			
			$this->booking_id = $booking_id;
			
			$sql = $wpdb->prepare("SELECT * FROM ".$wpdb->prefix."wpsg_ticket_booking WHERE id = %d", $booking_id);
			$this->data = $wpdb->get_row($sql, ARRAY_A);
			
		}
		
		public function getId(): int { return $this->booking_id; }
		public function getOrderID(): int { return intval($this->data['order_id']); }
		public function getOrder(): \wpsg_order { return \wpsg_order::getInstance($this->getOrderID()); }
		public function getOrderProductID(): int { return intval($this->data['order_product_id']); }
		public function getOrderProduct(): \wpsg_order_product { return \wpsg_order_product::getInstance($this->getOrderProductID()); }
		public function getProductID(): int { return intval($this->data['product_id']); }
		public function getProduct(): TicketProduct { return TicketProduct::getInstance($this->getProductID()); }
		public function getTicketID(): int { return intval($this->data['ticket_id']); }
		public function getTicket(): Ticket { return Ticket::getInstance($this->getTicketID()); }
		public function getAmountIndex(): int { return intval($this->data['amount_index']); }
		public function getBookingKey(): string { return $this->data['booking_key']; }
		public function isRevoked(): bool { return intval($this->data['revoke_user']??0) > 0; }
		public function isSend(): bool { return intval($this->data['send_user']) > 0; }
		public function getStatusCode(): int {
			
			if ($this->isRevoked()) return self::STATUS_UNAVAILABLE;
		
			$wpsg_mod_ticket_orderstate_active = Plugin::getInstance()->getActiveOrderState();
			
			if (!in_array($this->getOrder()->getStatusCode(), $wpsg_mod_ticket_orderstate_active)) return self::STATUS_UNAVAILABLE;
			
			return self::STATUS_AVAILABLE;
		
		}
		
		public function getRevokeTooltip(): string {
			
			if ($this->isRevoked()) {
				
				$tSend = strtotime($this->data['revoke_time']);
				
				return wpsg_translate(__('Am #1# um #2# von #3# entwertet.', 'wpsg'), date('d.m.Y', $tSend), date('H:i', $tSend), \get_userdata(intval($this->data['revoke_user']))->display_name);
				
			} else return '';
			
		}
		
		public function isSendTooltip(): string {
			
			if ($this->isSend()) {
				
				$tSend = strtotime($this->data['send_time']);
				
				return wpsg_translate(__('Am #1# um #2# von #3# versendet.', 'wpsg'), date('d.m.Y', $tSend), date('H:i', $tSend), \get_userdata(intval($this->data['send_user']))->display_name);
				
			} else return '';
			
		}
		
		public function getStateLabel(): string {
			
			return match($this->getStatusCode()) {
				self::STATUS_UNAVAILABLE => __('Inaktiv', 'wpsg'),
				self::STATUS_AVAILABLE => __('Aktiv', 'wpsg')
			};
		}
		
		public function isTimeRaised(): bool {
			
			return !$this->getTicket()->isActive(new \DateTime());
			
		}
		
		public function revoke(): void {
		
			global $wpdb;
			
			$result = $wpdb->update($wpdb->prefix."wpsg_ticket_booking", [
				'revoke_user' => \get_current_user_id(),
				'revoke_time' => \current_time('mysql')
			], [
				'id' => $this->getId()
			]);
			
			$this->addBookingLog(self::LOG_REVOKE);
		
		}
		
		public function setSend(): void {
			
			global $wpdb;
			
			$wpdb->update($wpdb->prefix."wpsg_ticket_booking", [
				'send_user' => \get_current_user_id(),
				'send_time' => \current_time('mysql')
			], [
				'id' => $this->getId()
			]);
			
		}
		
		public function getBookingLog(): array {
			
			global $wpdb;
			
			$query = $wpdb->prepare(
				"SELECT * FROM ".$wpdb->prefix."wpsg_ticket_log  WHERE booking_id = %d",
				$this->getId()
			);
			
			$arReturn = [];

			foreach ($wpdb->get_results($query, ARRAY_A) as $k => $l) {
				
				$l['log'] = intval($l['log']);
				
				$user_data = \get_userdata(intval($l['user_id']));
				
				$l['user_name'] = $user_data->display_name;
				
				$arReturn[] = $l;
				
			}
			
			return $arReturn;
			
		}
		
		public function toJSON(bool $bLoadSibling = true): array {
			
			$oOrderProduct = \wpsg_order_product::getInstance($this->getOrderProductID());
			$oTicket = Ticket::getInstance($this->getTicketID());
			$oShop = \wpsg_ShopController::getShop();
			
			$revoke_text = '';
			
			if ($this->isRevoked()) {
				
				$revoke_user = \get_userdata(intval(intval($this->data['revoke_user'])));
				$revoke_time = strtotime($this->data['revoke_time']);
				
				$revoke_text = wpsg_translate(__('Am #1# um #2# von #3# entwertet.', 'wpsg'), date('d.m.Y', $revoke_time), date('H:i', $revoke_time), $revoke_user->display_name);
				
			} else {
				
				if (!$this->getTicket()->isActive(new \DateTime())) $revoke_text = __('Ausgelaufen', 'wpsg');
				else {
					
					$wpsg_mod_ticket_orderstate_active = Plugin::getInstance()->getActiveOrderState();
					if (!in_array($this->getOrder()->getStatusCode(), $wpsg_mod_ticket_orderstate_active)) $revoke_text = __('Bestellstatus ungültig', 'wpsg');
				
				}
				
			}
			
			$r = [
				'label' => $this->getLabel(),
				'booking_id' => $this->getId(),
				'ticket_product_label' => $this->getProduct()->getProduct()->getProductName(),
				'order_product_price' => $this->getOrderProduct()->getPrice(\wpsg_ShopController::getShop()->getBackendTaxview()),
				'order_product_price_ff' => wpsg_ff($this->getOrderProduct()->getPrice($oShop->getBackendTaxview()), $oShop->get_option('wpsg_currency')),
				'order_product_id' => $this->getOrderProductID(),
				'amount_index' => $this->getAmountIndex(),
				'revoked' => $this->isRevoked(),
				'ticket_id' => $oTicket->getId(),
				'ticket_data' => $oTicket->toJSON(),
				'booking_key' => $this->getBookingKey(),
				'order_id' => $this->getOrderID(),
				'order_nr' => $this->getOrder()->getNr(),
				'order_state_label' => $this->getOrder()->getStateLabel(),
				'order_state' => $this->getOrder()->getStatusCode(),
				'wpsg_mod_ticket_orderstate_active' => Plugin::getInstance()->getActiveOrderState(),
				'order_title' => $this->getOrder()->getInvoiceTitle(),
				'order_firstname' => $this->getOrder()->getInvoiceFirstName(),
				'order_lastname' => $this->getOrder()->getInvoiceName(),
				'time_raised' => $this->isTimeRaised(),
				'state_code' => $this->getStatusCode(),
				'state_label' => $this->getStateLabel(),
				'log' => $this->getBookingLog(),
				'is_send' => $this->isSend(),
				'is_send_tooltip' => $this->isSendTooltip(),
				'revoke_text' => $revoke_text,
				'urls' => [
					'pdf' => $this->getTicketPublicURL(),
					'order_edit' => wpsg_admin_url('Order', 'view', ['edit_id' => $this->getOrderID()], [], true),
					'product_edit' => html_entity_decode(wp_nonce_url(WPSG_URL_WP.'wp-admin/admin.php?page=wpsg-Produkt&action=edit&edit_id='.$this->getProductID(), 'wpsg-product-edit-'.$this->getProductID()))
				]
			];
			
			$arProductVarData = \wpsg_ShopController::getShop()->callMod('wpsg_mod_productvars', 'getAllProductVarValues', [$this->getOrderID(), $this->getOrderProduct()->getProductIndex()]);
			
			$r['arProductVars'] = [];
			
			if (sizeof($arProductVarData??[]) > 0) {
				
				$r['arProductVars'] = array_values($arProductVarData);
				
			}
			
			
			if ($bLoadSibling) $r['sibling'] = $this->getSiblingTicketBooking();
			
			return $r;
			
		}
		
		public function getLabel(): string  {
			
			return 'Ticket #'.$this->getOrderProductID().'-'.($this->getAmountIndex() + 1);
			
		}
		
		public function getTicketPublicURL(): string { return $this->getTicket()->getTicketPublicURL($this->getOrderProductID(), $this->getAmountIndex()); }
		
		/**
		 * Gibt weitere TicketBuchungen
		 * @return array
		 */
		public function getSiblingTicketBooking(): array {
		
			$arTicketBooking = [];
			
			foreach ($this->getOrder()->getOrderProducts() as $oOrderProduct) {
				
				$ticket_id = intval($oOrderProduct->getMeta('ticket_id'));
				
				if ($ticket_id > 0) {
					
					for ($amount_index = 0; $amount_index < $oOrderProduct->getAmount(); $amount_index ++) {
						
						// if ($this->getOrderProductID() !== $oOrderProduct->getId() || $this->getAmountIndex() !== $amount_index) {
							
						$oTicketBooking = TicketBooking::getInstanceByBookingKey(TicketBooking::encodeBookingKey(
							$oOrderProduct->getId(), $amount_index
						));
					
						$arTicketBooking[] = $oTicketBooking->toJSON(false);
							
						//}
						
					}
					
				}
				
			}
			
			return $arTicketBooking;
			
		}
		
		public function addBookingLog(int $log): void {
			
			global $wpdb;
			
			$wpdb->insert(
				$wpdb->prefix."wpsg_ticket_log",
				[
					'booking_id' => $this->getId(),
					'log' => $log,
					'user_id' => \get_current_user_id(),
					'date_time' => \current_time('mysql')
				]
			);
		
		}
		
		public function getTicketPDFPath(): string {
			
			return $this->getStoragePath().'ticket_'.$this->getId().'.pdf';
			
		}
		
		public function getStoragePath(): string {
			
			 if (\wpsg_ShopController::getShop()->isMultiBlog()) {

                $path = WP_CONTENT_DIR.'/'.WPSG_MB_UPLOADS.'/wpsg/wpsg_ticket/tickets/';

            } else {

                $path = WP_CONTENT_DIR.'/uploads/wpsg/wpsg_ticket/tickets/';

            }
			 
			$path .= $this->getId().DIRECTORY_SEPARATOR;
			
			\wpsg_ShopController::getShop()->protectDirectory($path);
			 
			 return $path;
			
		}
		
		/** Static */
		
		public static function encodeBookingKey(int $order_product_id, int $amount_index): string {
			
			$oOrderProduct = \wpsg_order_product::getInstance($order_product_id);
			
			$ticket_id = $oOrderProduct->getMeta('ticket_id');
			
			return $oOrderProduct->getProductId().'|'.$oOrderProduct->getOrder()->getId().'|'.$order_product_id.'|'.$ticket_id.'|'.$amount_index;
			
		}
		
		public static function decodeBookingKey(string $booking_key): array {
			
			$arKey = explode('|', $booking_key);
			
			return [
				'product_id' => intval($arKey[0]),
				'order_id' => intval($arKey[1]),
				'order_product_id' => intval($arKey[2]),
				'ticket_id' => intval($arKey[3]),
				'amount_index' => intval($arKey[4])
			];
			
		}
		
		public static function getInstanceByBookingKey(string $booking_key) {
			
			global $wpdb;
			
			$id = intval($wpdb->get_var($wpdb->prepare("SELECT `id` FROM ".$wpdb->prefix."wpsg_ticket_booking WHERE booking_key = '%s' ", $booking_key)));
			
			if ($id <= 0) {
				
				list($product_id, $order_id, $order_product_id, $ticket_id, $amount_index) = array_values(self::decodeBookingKey($booking_key));
				
				$wpdb->insert($wpdb->prefix."wpsg_ticket_booking", [
					'booking_key' => $booking_key,
					'product_id' => $product_id,
					'order_id' => $order_id,
					'order_product_id' => $order_product_id,
					'ticket_id' => $ticket_id,
					'amount_index' => $amount_index
				]);
				
				$id = $wpdb->insert_id;
				
			}
			
			return self::getInstance($id);
			
		}
		
		public static function getInstance(int $id, bool $no_cache = false) {
			
			if (!isset(self::$arInstances[$id]) || $no_cache === true) self::$arInstances[$id] = new self($id);
			
			return self::$arInstances[$id];
			
		}
	
	}